﻿module ExtendedLambdaParser

open FParsec
open LambdaCalculus

type ExTerm =   
    Variable      of string
   | Lambda       of string list * ExTerm 
   | Application  of ExTerm * ExTerm
   | LetIn        of string * ExTerm * ExTerm
   | Substitution of ExTerm * string * ExTerm
   | Delay        of ExTerm
   | Literal      of int

let ws = spaces .>> opt (between (pstring "%") (pstring "%") (manySatisfy (isNoneOf "%"))) >>. spaces
let strws st = pstring st .>> ws

let parsec = new ParserCombinator()

let identifier = attempt (parsec {
        let fstDigitParser = isNoneOf ".\()[]%${}0123456789 \t\r\n"
        let! str = many1Satisfy2L fstDigitParser (isNoneOf ".\()[]${} \t\r\n") "identifier" .>> ws
        if [ "let"; "in" ] |> Set.ofList |> Set.contains str then
            return! fail("Sensitive keyword")
        else
            return str
    })

let variable = identifier |>> Variable
let number   = many1Satisfy isDigit .>> ws |>> (fun s -> Literal(int s))

let term, termref = createParserForwardedToRef<ExTerm, unit>()

let lambda = (strws "\\" >>. many1 identifier .>> strws ".") .>>. ((term .>> ws) <?> "function body")
            |>> Lambda

let atom = number <|> variable <|> between (strws "(") (strws ")") term <|> between (strws "{") (strws "}") (term |>> Delay)
let atomOrSubs = atom .>>. (opt (between (strws "[") (strws "]") (identifier .>>. (strws ":=" >>. term))) <?> "substitution clause")
                 |>> (function
                        | term, None -> term
                        | term, Some(x, expn) -> Substitution(term, x, expn)
                     )
let applications = many1 atomOrSubs |>> List.reduce (fun a b -> Application(a,b))

let letStm = pipe3 (strws "let" >>. identifier .>> strws "=") (applications .>> strws "in") term (fun var value body -> LetIn(var, value, body))

termref := (letStm <|> lambda <|> applications) <?> "term"

let fullTerm = ws >>. term .>> eof

let rec rewrite exterm = 
    match exterm with
    | Variable(s) -> LambdaCalculus.Variable(s)
    | Application(a, b) -> LambdaCalculus.Application(rewrite a, rewrite b)
    | Lambda(vars, body) -> List.fold (fun expn var -> LambdaCalculus.Lambda(var, expn)) (rewrite body) (List.rev vars)
    | LetIn(x, expn, body) -> LambdaCalculus.Application(LambdaCalculus.Lambda(x, rewrite body), rewrite expn)
    | Substitution(term, var, expn) -> rewrite (LetIn(var, expn, term))
    | Delay(term) -> LambdaCalculus.Lambda("$", rewrite term)
    | Literal(i) -> LambdaCalculus.Lambda("f", LambdaCalculus.Lambda("x", List.fold(fun exp _ -> LambdaCalculus.Application(LambdaCalculus.Variable("f"), exp)) (LambdaCalculus.Variable("x")) [ 1..i ] ))